
  class MailScanner < Scanner

    # mode

    MODE_NONE = :MODE_NONE

    MODE_822  = :MODE_822
    MODE_MIME = :MODE_MIME

    MODE_RECV = :MODE_RECV


    def reset( str, fld )
      super( str )
      @header = fld

      @m_atom = MODE_822
      @m_recv = MODE_NONE

      case fld
      when CTypeH
        @m_atom = MODE_MIME
      when CEncodingH, CDispositionH
        @m_atom = MODE_MIME
      when RecvH
        @m_recv = MODE_RECV
      end
    end


    # exp

    atomchars  = Regexp.quote( "\#!$%&`'*+{|}~^/=?" ) + '\\-'
    tokenchars = Regexp.quote( "\#!$%&`'*+{|}~^." )   + '\\-'
    eucchars   = "\xa1\xa1-\xf3\xfe"
    jisstr     = "\\e..[^\\e]*\\e.."

    ATOM    = /\A[\w#{atomchars}#{eucchars}]+|#{jisstr}/o
    TOKEN   = /\A[\w#{tokenchars}#{eucchars}]+|#{jisstr}/o
    DIGIT   = /\A\d+\z/o

    LWSP    = /\A(?:\n|\r\n|\r)?[ \t]+/o

    BACKSLASH = /\A\\/o

    BEGIN_Q = /\A"/o
    Q_ENT   = /\A[^"\\\e]+|#{jisstr}/o
    END_Q   = /\A"/o

    BEGIN_C = /\A\(/o
    C_ENT   = /\A[^\)\(\\\e]+|#{jisstr}/o
    END_C   = /\A\)/o

    BEGIN_D = /\A\[/o
    D_ENT   = /\A[^\]\\]+|#{jisstr}/o
    END_D   = /\A\]/o


    OMIT = '$omit'

    def scan( cut = true )
      sret = nil
      vret = OMIT

      until sret do
        unless @scan.rest? then
          @spipe.push false ; @vpipe.push nil
          @spipe.push false ; @vpipe.push nil
          break
        end

        @scan.skip LWSP

        if @m_atom == MODE_822 then
          if vret = @scan.fullscan( ATOM, true, cut ) then
            sret = :ATOM

            if DIGIT === vret then
              sret = :DIGIT
              vret = vret.to_i

            elsif @m_recv = MODE_RECV then
              case vret.downcase
              when 'from' then sret = :FROM
              when 'by'   then sret = :BY
              when 'via'  then sret = :VIA
              when 'with' then sret = :WITH
              when 'id'   then sret = :ID
              when 'for'  then sret = :FOR
              end
            end
            break
          end

        elsif @m_atom == MODE_MIME then
          if vret = @scan.fullscan( TOKEN, cut, cut ) then
            sret = :TOKEN
            break
          end
        else
          bug! 'atom mode is not 822/MIME'
        end

        if @scan.fullscan( BEGIN_Q, cut, cut ) then
          sret = :QUOTED
          vret = quoted if cut

        elsif @scan.skip( BEGIN_C ) then
          @header.comments.push comment

        elsif @scan.fullscan( BEGIN_D, cut, cut ) then
          sret = :DOMLIT
          vret = domlit if cut

        else
          sret = vret = @scan.getch
          unless cut then
            @scan.unscan
          end
        end
      end

      if cut then
        @spipe.push sret
        @vpipe.push vret
      end
      report_ret( sret, vret ) if @debug

      return sret
    end



    private


    def quoted
      ret = ''
      while true do
        if    temp = @scan.scan( Q_ENT ) then ret << temp
        elsif @scan.skip( END_Q )        then break
        elsif @scan.skip( BACKSLASH )    then ret << @scan.getch
        else
          unless @scan.rest? then
            scan_error! "found unterminated quoted-string"
          end
          bug! 'in quoted, no match'
        end
      end

      return ret
    end

    
    def comment
      ret = ''
      nest = 1

      while nest > 0 and @scan.rest? do
        if    temp = @scan.scan( C_ENT ) then ret << temp
        elsif @scan.skip( END_C )        then nest -= 1
        elsif @scan.skip( BEGIN_C )      then nest += 1
        elsif @scan.skip( BACKSLASH )    then ret << @scan.getch
        else
          unless @scan.rest? then
            scan_error! "found unterminated comment"
          end
          bug! 'in comment, no match'
        end
      end
      if nest > 0 then
        scan_error! "found unterminated comment"
      end

      return ret
    end

    
    def domlit
      ret = ''

      while true do
        if temp = @scan.scan( D_ENT ) then
          ret << temp
        end
        if    @scan.skip( END_D )     then break
        elsif @scan.skip( BACKSLASH ) then ret << @scan.getch
        else
          unless @scan.rest? then
            scan_error! "found unterminated domain literal"
          end
          bug! 'in domlit, no match'
        end
      end

      return ret
    end


    def report_ret( sret, ret )
      if Integer === sret then
        puts "sret #{sret.id2name || sret}"
      else
        puts "sret #{sret.inspect}"
      end
      puts "vret #{ret}"
    end

  end
